module ComputationM
  ( ComputationM
  , evalComputationM
  , inject
  , mapComputationM
  , runFunction
  , runFunctionIO
  ) where

import Types
import Sources

import Control.Monad.State
import Control.Monad.Writer

data TypedRun a b where
  Function :: IsFunction f a b => f -> TypedRun a b
  FunctionIO :: IsFunctionIO f a b => f -> TypedRun a b
  Inject :: b -> TypedRun () b
  GetListElem :: TypedRun [b] b
  SetListElem :: TypedRun a [a]
  MapComp :: [DependencyUntyped] -> TypedRun [a] [b]

deriving instance (Show a, Show b) => Show (TypedRun a b)

data Token a where
  Token :: Int -> Token a
  NoToken :: Token ()

deriving instance Show (Token a)

data Dependency a b = Dependency (Token a) (TypedRun a b) (Token b)
  deriving (Show)

data DependencyUntyped where
  DependencyUntyped :: Dependency a b -> String -> DependencyUntyped

instance Show DependencyUntyped where
  show (DependencyUntyped _ s) = s

newtype ComputationM' a = ComputationM { unComputationM :: WriterT [DependencyUntyped] (State Int) a }
  deriving (Functor, Applicative, Monad, MonadState Int, MonadWriter [DependencyUntyped])

type ComputationM a = ComputationM' (Token a)

evalComputationM' :: Int -> ComputationM () -> ([DependencyUntyped], Int)
evalComputationM' top m = runState (execWriterT (unComputationM m)) top

evalComputationM :: ComputationM () -> [DependencyUntyped]
evalComputationM m = fst (evalComputationM' 0 m)

tellDep :: (Show a, Show b) => Dependency a b -> ComputationM' ()
tellDep dep = tell [DependencyUntyped dep (show dep)] -- Call show for later debugging purposes

genDependency' :: (Show u, Show a) => (Token a -> ComputationM' (Dependency u a)) -> ComputationM a
genDependency' f = do
  top <- get
  let top' = top + 1
      target = Token top'
  put top'
  result <- f target
  tellDep result
  pure target

genDependency :: (Show u, Show a) => (Token a -> Dependency u a) -> ComputationM a
genDependency f = genDependency' (pure . f)

inject :: Show a => a -> ComputationM a
inject x = genDependency (Dependency NoToken (Inject x))

getListElem :: Show a => Token [a] -> ComputationM a
getListElem outer = genDependency (Dependency outer GetListElem)

setListElem :: Show a => Token a -> Token [a] -> ComputationM ()
setListElem a outer = do
  tellDep (Dependency a SetListElem outer)
  pure NoToken

runFunction :: (Show a, Show b, IsFunction f a b) => f -> Token a -> ComputationM b
runFunction f input = genDependency (Dependency input (Function f))

runFunctionIO :: (Show a, Show b, IsFunctionIO f a b) => f -> Token a -> ComputationM b
runFunctionIO f input = genDependency (Dependency input (FunctionIO f))

mapComputationM :: (Show a, Show b) => (Token a -> ComputationM b) -> Token [a] -> ComputationM [b]
mapComputationM f input = genDependency' $ \target -> do
  top <- get
  let (res, top') = evalComputationM' top $ do
        inp <- getListElem input
        outp <- f inp
        setListElem outp target
  put top'
  pure (Dependency input (MapComp res) target)
